require("config")
require("stdlib.util")

prioritize = function(loco)
--[[ Give locomotive priority so that it updates on every tick ]]
	global.high_prio_loco[loco.unit_number] = loco
end

deprioritize = function(loco)
--[[ Remove locomotive priority ]]
	global.high_prio_loco[loco.unit_number] = nil
end

update_pump = function(pump)
--[[ Update pump so they snap to the correct rail ]]
	pump.direction = oppositedirection(pump.direction)
	pump.direction = oppositedirection(pump.direction)
end

update_pumps_around_loco = function(loco)
--[[ Update pumps around locomotive so they snap to the correct rail ]]
	local pos = loco.position
	local pumps = loco.surface.find_entities_filtered{type = "pump", area = {{pos.x-5, pos.y-5}, {pos.x+5, pos.y+5}}}
	for _, pump in pairs(pumps) do
		update_pump(pump)
	end
end

update_loco_fuel = function(loco)
--[[ Update locomotive remaining fuel depends on amount of fluid in proxy_tank 
	if no proxy_tank found (update fail) return -1
	else return number of tick since fluid amount in proxy_tank last changed ]]
	local proxy = global.proxies[loco.unit_number]
	if not (proxy and proxy.tank.valid) then
		return (-1)
	end
	local burner_inventory = loco.burner.inventory
	local fluidbox = proxy.tank.fluidbox
	local fluid = fluidbox[1] or fluidbox[2] or fluidbox[3]
	local fluid_type = fluid and fluid.type
	if not (fluid_type and USABLE_FLUID[fluid_type]) then
		burner_inventory.clear()
		return (proxy.last_amount == 0) and (game.tick - proxy.tick) or 0
	end
	local fluid_amount = (
		(fluidbox[1] and fluidbox[1].amount or 0) +
		(fluidbox[2] and fluidbox[2].amount or 0) +
		(fluidbox[2] and fluidbox[2].amount or 0)
	)
	if proxy.last_amount == fluid_amount then
		return game.tick - proxy.tick
	end
	burner_inventory.clear()
	local amount = round(fluid_amount)
	if amount > 0 then
		burner_inventory.insert({name = fluid_type,	count = round(fluid_amount)})
	end
	proxy.tick = game.tick
	proxy.last_amount = fluid_amount
	return 0
end

create_proxy = function(loco)
--[[ Create proxy_tank for a locomotive and inserting the proxy_tank to global.proxies 
	if proxy_tank successfully created return 0, else return -1 ]]
	local uid = loco.unit_number
	local proxy = global.proxies[uid]
	if not(proxy and proxy.tank.valid) and math.floor(4 * loco.orientation) == 4 * loco.orientation then
		proxy_rail = {}
		for i = 1, 3 do
			proxy_rail[i] = loco.surface.create_entity{
				name = "proxy-straight-rail",
				position = moveposition(loco.position, ori_to_dir(loco.orientation), 2 * (i - 2)),
				force = loco.force,
				direction = ori_to_dir(loco.orientation)
			}
		end
		proxy_tank = loco.surface.create_entity{
			name = "proxy-tank",
			position = loco.position,
			force = loco.force,
			direction = ori_to_dir(loco.orientation)
		}
		if (not proxy_tank) then return -1 end
		local burner_inventory = loco.burner.inventory
		local fluid_amount = 0
		if burner_inventory[1] and burner_inventory[1].valid_for_read then
			local fluid_type = burner_inventory[1].name
			fluid_amount = burner_inventory[1].count/3
			if USABLE_FLUID[fluid_type] then
				for i = 1, 3 do proxy_tank.fluidbox[i] = {type = fluid_type, amount = fluid_amount} end
			end
		end
		global.proxies[uid] = {tank = proxy_tank, last_amount = 3 * fluid_amount, tick = game.tick, rail = proxy_rail}
		local update_tick = uid % SLOW_UPDATE_TICK + 1
		global.update_tick[uid] = update_tick
		global.low_prio_loco[update_tick][uid] = loco
		global.high_prio_loco[uid] = loco
		update_pumps_around_loco(loco)
		return 0
	end
	return -1
end

destroy_proxy = function(loco)
--[[ Update the locomotive then destroy the proxy_tank
	return number of ticks since last fluid change in proxy_tank
	return -1 if locomotive has no proxy_tank ]]
	local uid = loco.unit_number
	local no_update_ticks = update_loco_fuel(loco)
	if no_update_ticks >= 0 then
		local proxy_tank = global.proxies[uid].tank
		proxy_tank.destroy()
		for _,r in pairs(global.proxies[uid].rail) do
			r.destroy()
		end
		update_pumps_around_loco(loco)
		global.low_prio_loco[global.update_tick[uid]][uid] = nil
	end
	global.proxies[uid] = nil
	global.update_tick[uid] = nil
	global.high_prio_loco[uid] = nil
	return no_update_ticks
end

update_loco = function(loco)
--[[ If locomotive is idle, the fuel will be updated or a proxy_tank will be created
	if the locomotive is moving, proxy_tank will be destroyed
	Also put the locomotive to its appropriate priority ]]
	if loco.train.speed == 0 then
		local no_update_ticks = update_loco_fuel(loco)
		if no_update_ticks == -1 then
			create_proxy(loco)
		elseif no_update_ticks <= IDLE_TICK_BUFFER then
			prioritize(loco)
		else
			deprioritize(loco)
		end
	else
		destroy_proxy(loco)
	end
end

switch_loco = function(loco)
--[[ For updating from vanilla locomotive ]]
	local inventory = loco.burner.inventory
	local replacement_fuel = "crude-oil"
	local energy = loco.burner.remaining_burning_fuel/game.item_prototypes[replacement_fuel].fuel_value
	for i = 1, #inventory do
		if inventory[i] and inventory[i].valid_for_read then
			energy = energy + inventory[i].prototype.fuel_value/game.item_prototypes[replacement_fuel].fuel_value * inventory[i].count
			loco.remove_item({name = inventory[i].name, count = inventory[i].count})
		end
	end
	--game.players[1].print(#inventory)
	local result = math.min(math.floor(energy), TANK_CAPACITY)
	loco.insert({name = replacement_fuel, count = math.max(result, 1)})
end

train_ridden = function()
--[[ Return array of trains that is in manual_control and ridden by a player]]
	local trains = {}
	for i,p in pairs(game.players) do
		if (
			p.vehicle and
			(
				p.vehicle.type == "fluid-wagon" or 
				p.vehicle.type == "cargo-wagon" or
				p.vehicle.type == "locomotive"
			) and
			p.vehicle.train.state == defines.train_state.manual_control
		) then
			trains[i] = p.vehicle.train
		end
	end
	return trains
end

update_train = function(train)
--[[ Update all locomotives in train ]]
	for _,l in pairs(train.locomotives.front_movers) do
		update_loco(l)
	end
	for _,l in pairs(train.locomotives.back_movers) do
		update_loco(l)
	end
end

create_proxy_rail = function(rail)
--[[ Create proxy rail for proxy_tank placement ]]
	local proxy_rail = rail.surface.create_entity{
		--name = ("proxy-"..rail.name),
		name = "proxy-straight-rail",
		position = rail.position,
		force = rail.force,
		direction = rail.direction
	}
	game.players[1].print(proxy_rail == nil)
end

destroy_proxy_rail = function(rail)
--[[ Destroy proxy rail ]] 
	local box = rail.bounding_box
	local proxies = rail.surface.find_entities_filtered{name = ("proxy-"..rail.name), area = box}
	if next(proxies) ~= nil then
		for _,r in pairs(proxies) do
			if r.direction == rail.direction then
				r.destroy()
			end
		end
	end
end

ON_BUILT = function(event)
--[[ Handler for when entity is built ]]
	local entity = event.created_entity
	local surface = entity.surface
	local name = entity.name
	local position = entity.position
	local force = entity.force
	local direction = entity.direction
	if entity.name == "locomotive" then
		update_loco(entity)
	end
end

ON_DESTROYED = function(event)
--[[ Handler for when entity is destroyed ]]
	local entity = event.entity 
	if (entity.name == "straight-rail") or (entity.name == "curved-rail") then
		destroy_proxy_rail(entity)
	elseif (entity.type == "locomotive") then
		destroy_proxy(entity)
	end
end

ON_TICK = function(event)
--[[ Handler for every tick ]]
	if TICK_UPDATE then
		for _, l in pairs(global.low_prio_loco[event.tick % SLOW_UPDATE_TICK + 1]) do
			update_loco(l)
		end
		for _, l in pairs(global.high_prio_loco) do
			update_loco(l)
		end
	end
	for _,t in pairs(train_ridden()) do
			update_train(t)
	end
end

ON_TRAIN_CHANGED_STATE = function (event)
--[[ Handler for when a train changed state ]]
	local train = event.train
	local state = train.state
	local train_state = defines.train_state
	local stopped = (state == (train_state.no_schedule)) or (state == (train_state.no_path)) or (state == (train_state.wait_station)) or (state == (train_state.manual_control))
	if not (state == train_state.wait_signal) then update_train(train) end
	if not stopped then
		for _,loco in pairs(train.locomotives.front_movers) do
			destroy_proxy(loco)
		end
		for _,loco in pairs(train.locomotives.back_movers) do
			destroy_proxy(loco)
		end
	end
end

ON_PLAYER_DRIVING_CHANGED_STATE = function(event)
--[[ Handler for when player enter or exit vehicle ]]
	local player = game.players[event.player_index]
	local vehicle = player.vehicle
	if vehicle ~= nil and vehicle.name == "proxy-tank" then
		local correct_vehicle = vehicle.surface.find_entities_filtered{name = "locomotive", position = vehicle.position}[1]
		if correct_vehicle then
			correct_vehicle.passenger = player
		end
	end
end

is_fake_item = function(item_stack)
--[[ Return true if the item_stack is a proxy item created by the mod ]]
	return (item_stack.prototype.fuel_category == "fluid")
end

ON_PLAYER_CURSOR_STACK_CHANGED = function(event)
--[[ Handler for when cursor pick up or put down something ]]
	local player = game.players[event.player_index]
	local taken_item = player.cursor_stack
	if taken_item and taken_item.valid_for_read and is_fake_item(taken_item) and player.opened and player.opened.name == "locomotive" then
		player.opened.insert(taken_item)
		taken_item.clear()
	end
end

ON_PLAYER_MAIN_INVENTORY_CHANGED = function(event)
--[[ Handler for when player main inventory changed ]]
	local player = game.players[event.player_index]
	local inventory = player.get_inventory(defines.inventory.player_main)
	for name, item in pairs(USABLE_FLUID) do
		if game.item_prototypes[name] then
			local proxy_name = name
			local amount = inventory.remove({name = proxy_name, count = TANK_CAPACITY})
			if amount and amount > 0 and player.opened and player.opened.name == "locomotive" then
				player.opened.insert({name = proxy_name, count = amount})
			end
		end
	end
	
end

script.on_event({defines.events.on_built_entity, defines.events.on_robot_built_entity}, ON_BUILT)
script.on_event({defines.events.on_player_mined_entity, defines.events.on_entity_died, defines.events.on_robot_mined_entity}, ON_DESTROYED)
script.on_event({defines.events.on_tick}, ON_TICK)
script.on_event({defines.events.on_train_changed_state}, ON_TRAIN_CHANGED_STATE)
script.on_event({defines.events.on_player_driving_changed_state}, ON_PLAYER_DRIVING_CHANGED_STATE)
script.on_event({defines.events.on_player_cursor_stack_changed}, ON_PLAYER_CURSOR_STACK_CHANGED)
script.on_event({defines.events.on_player_main_inventory_changed}, ON_PLAYER_MAIN_INVENTORY_CHANGED)

ON_INIT = function()
--[[ Handler for when the mod is loaded the first time into a save ]]
	global = global or {}
	global.proxies = global.proxies or {} -- contain proxy_tank for each locomotive and its functionality
	global.update_tick = global.update_tick or {} -- contain locomotive entity that have proxy and on what tick
	global.low_prio_loco = global.low_prio_loco or {} -- contain on what ticks loco are updated slowly
	for i=1,SLOW_UPDATE_TICK do
		global.low_prio_loco[i] = {}
	end
	global.high_prio_loco = global.high_prio_loco or {} -- contain locomotives that updates every tick (priority_loco will be subset of locomotives)
	for _, force in pairs(game.forces) do
			for _, train in pairs(force.get_trains()) do
				for _, l in pairs(train.locomotives.front_movers) do
					if l.name == "locomotive" then
						switch_loco(l)
						update_loco(l)
					end
				end
				for _, l in pairs(train.locomotives.back_movers) do
					if l.name == "locomotive" then
						switch_loco(l)
						update_loco(l)
					end
				end
			end
		end
end

function create_gui(player)
	player.gui.top.add({type = "sprite-button", name = "Diesel_Locomotive_main", sprite = "entity/locomotive", style = "mod_gui_button_style"})
end

function toggle_gui(player)
	if player.gui.left.Diesel_Locomotive_frame then
		player.gui.left.Diesel_Locomotive_frame.destroy()
	else
		player.gui.left.add({type = "frame", name = "Diesel_Locomotive_frame", style = "frame_style", caption = "Diesel Locomotive Options"})
		player.gui.left.Diesel_Locomotive_frame.add{type = "button", name = "Diesel_Locomotive_uninstall", style = "button_style", caption = "Uninstall"}
		player.gui.left.Diesel_Locomotive_frame.add{type = "button", name = "Diesel_Locomotive_reinstall", style = "button_style", caption = "Reinstall"}
	end
end

function ON_PLAYER_JOINED_GAME(event)
	local player = game.players[event.player_index]
	create_gui(player)
end

function ON_GUI_CLICK(event)
	local player = game.players[event.player_index]
	if event.element.name == "Diesel_Locomotive_main" then
		toggle_gui(player)
	elseif event.element.name == "Diesel_Locomotive_uninstall" then
		script.on_event(
			{
				defines.events.on_built_entity,
				defines.events.on_robot_built_entity,
				defines.events.on_player_mined_entity,
				defines.events.on_entity_died,
				defines.events.on_robot_mined_entity,
				defines.events.on_tick,
				defines.events.on_train_changed_state,
				defines.events.on_player_driving_changed_state
			},
			function() end
		)
		for _, force in pairs(game.forces) do
			for _, train in pairs(force.get_trains()) do
				for _, l in pairs(train.locomotives.front_movers) do
					if l.name == "locomotive" then
						destroy_proxy(l)
					end
				end
				for _, l in pairs(train.locomotives.back_movers) do
					if l.name == "locomotive" then
						destroy_proxy(l)
					end
				end
			end
		end
		player.print("Diesel locomotive uninstalled, locomotives no longer accept fluid (also don't accept solid fuel either).")
		player.print("Keep in mind that if you disable the mod after saving the fluid fuels inside locomotives will be deleted also.")
		player.print("Press reinstall to re-enable locomotives to receive fluid")
	elseif event.element.name == "Diesel_Locomotive_reinstall" then
		script.on_event({defines.events.on_built_entity, defines.events.on_robot_built_entity}, ON_BUILT)
		script.on_event({defines.events.on_player_mined_entity, defines.events.on_entity_died, defines.events.on_robot_mined_entity}, ON_DESTROYED)
		script.on_event({defines.events.on_tick}, ON_TICK)
		script.on_event({defines.events.on_train_changed_state}, ON_TRAIN_CHANGED_STATE)
		script.on_event({defines.events.on_player_driving_changed_state}, ON_PLAYER_DRIVING_CHANGED_STATE)
		for _, force in pairs(game.forces) do
			for _, train in pairs(force.get_trains()) do
				for _, l in pairs(train.locomotives.front_movers) do
					if l.name == "locomotive" then
						update_loco(l)
					end
				end
				for _, l in pairs(train.locomotives.back_movers) do
					if l.name == "locomotive" then
						update_loco(l)
					end
				end
			end
		end
		player.print("Diesel locomotive reinstalled, locomotives accept fluids once more.")
	end
end
script.on_init(ON_INIT)
script.on_event({defines.events.on_player_joined_game}, ON_PLAYER_JOINED_GAME)
script.on_event({defines.events.on_gui_click}, ON_GUI_CLICK)
